// MulticastCP.cpp : main project file.
//
// This sample illustrates IP multicasting with both the IPv4 and IPv6 protocol. This sample
// can act as both a sender and receiver of multicast data. If invoked as a receiver, the 
// specified multicast groups are joined on the given interface. Note that this sample
// only allows all groups to be joined on one specified interface but it is possible to
// have a socket joined to multiple multicast groups on different interfaces (with the
// socket being bound to the wildcard address). After the socket is created and joined to
// the multicast groups, it receives data destined to those groups. If invoked as a sender,
// the socket is created and the specified multicast groups are joined and the outgoing
// send interface is set to the local interface specified. Note that a multicast sender
// is not required to join any multicast groups to send data to it.
//
// usage:
//      Executable_file_name [/i ip] [/b ip] [/m ip] [/p port] [/x num] [/t text]
//                    [/n count] [/s | /r]
//         /i ip		Local interface to join multicast group(s) on
//         /b ip		Local IP to bind the socket to
//         /m ip		Multicast group to join
//         /p port	    Port to listen on or send to
//         /x size	    Length of send or receive buffer
//         /t text	    Text message to put in send buffer
//         /n count     Number of times to send or receive
//         /s           Send data to a multicast group
//         /r		    Receive data send to the group
//
// sample usage:
//      The following command line invokes a multicast receiver (where x.y.z.w is the
//      local interface to join the group on):
//          Executable_file_name -r -i x.y.z.w -m 234.6.6.6
//
//      The following command line invokes a multicast sender (where a.b.c.d is the local
//      interface to join the group on that is routable to x.y.z.w):
//          Executable_file_name -s -i a.b.c.d -m 234.6.6.6
//
#include "stdafx.h"

using namespace System;
using namespace System;
using namespace System::Collections;
using namespace System::Net;
using namespace System::Net::Sockets;


public ref class MulticastEndpoint
{
public:
	ArrayList^	localInterfaceList;
	ArrayList^	multicastJoinList;				  
	Socket^           mcastSocket;
	IPAddress^        bindAddress;
	int              bufferSize;
	int              localPort;    
	array<Byte>^     dataBuffer;

	/// <summary>
	/// A simple constructor 
	/// </summary>
	public:
	MulticastEndpoint()
	{
	   localInterfaceList = gcnew ArrayList();
	   multicastJoinList = gcnew ArrayList();
	   bufferSize = 512;
	   mcastSocket = nullptr;
	}

	/// <summary>
	/// This method creates the socket, joins it to the given multicast groups, and initializes
	/// the send/receive buffer. Note that the local bind address should be the wildcard address
	/// because it is possible to join multicast groups on one or more local interface and if
	/// a socket is bound to an explicit local interface, it can lead to user confusion (although
	/// this does currently work on Windows OSes).
	/// </summary>
	/// <param name="port">Local port to bind socket to</param>
	/// <param name="bufferLength">Legnth of the send/recv buffer to create</param>
public:
	void Create(int port, int bufferLength)
	{       
		localPort = port;

		Console::WriteLine("Creating socket, joining multicast group and");
		Console::WriteLine("     initializing the send/receive buffer");
		try
		{
			// If no bind address was specified, pick an appropriate one based on the multicast
			// group being joined.
			if ( bindAddress == nullptr )
			{
				IPAddress^ tmpAddr = (IPAddress^) multicastJoinList[0];

				if ( tmpAddr->AddressFamily == AddressFamily::InterNetwork )
					bindAddress = IPAddress::Any;
				else if ( tmpAddr->AddressFamily == AddressFamily::InterNetworkV6 )
					bindAddress = IPAddress::IPv6Any;
			}

			// Create the UDP socket
			Console::WriteLine("Creating the UDP socket...");
			mcastSocket = gcnew Socket(
				bindAddress->AddressFamily,                
				SocketType::Dgram,
				(ProtocolType)0
				);

			Console::WriteLine("{0} (IPv4 or IPv6) multicast socket created...", bindAddress->AddressFamily);

			// Bind the socket to the local endpoint
			Console::WriteLine("Binding the socket to the local endpoint...");
			IPEndPoint^  bindEndPoint = gcnew IPEndPoint( bindAddress, port );
			mcastSocket->Bind( bindEndPoint );
			Console::WriteLine("Multicast socket bound to: {0}", bindEndPoint->ToString());

			// Join the multicast group
			Console::WriteLine("Joining the multicast group...");
			for(int i = 0; i < multicastJoinList->Count ;i++)
			{
				for(int j = 0; j < localInterfaceList->Count ;j++)
				{
					// Create the MulticastOption structure which is required to join the 
					//    multicast group
					if ( mcastSocket->AddressFamily == AddressFamily::InterNetwork )
					{
						MulticastOption^     mcastOption = gcnew MulticastOption( 
							(IPAddress^) multicastJoinList[i], 
							(IPAddress^) localInterfaceList[j] 
							);
	            
						mcastSocket->SetSocketOption(
							SocketOptionLevel::IP,                    
							SocketOptionName::AddMembership,
							mcastOption
							);
					}
					else if ( mcastSocket->AddressFamily == AddressFamily::InterNetworkV6 )
					{
						IPv6MulticastOption^ ipv6McastOption = gcnew IPv6MulticastOption(
							(IPAddress^) multicastJoinList[i], Convert::ToInt64(localInterfaceList[j]));

						mcastSocket->SetSocketOption(
							SocketOptionLevel::IPv6,
							SocketOptionName::AddMembership,
							ipv6McastOption
							);
					}
	            
					Console::WriteLine("Joined multicast group {0} on interface {1}", 
						multicastJoinList[i]->ToString(),
						localInterfaceList[j]->ToString()
						);
				}
			}

			// Allocate the send and receive buffer
			Console::WriteLine("Allocating the send and receive buffer...");
			dataBuffer = gcnew array<Byte>( bufferLength );
		}
		catch(SocketException^ err)
		{
			Console::WriteLine("Exception occured when creating multicast socket: {0}", err->Message);
		}
	}

	/// <summary>
	/// This method drops membership to any joined groups. To do so, you have to
	/// drop the group exactly as you joined it -- that is the local interface
	/// and multicast group must be the same as when it was joined. Also note
	/// that it is not required to drop joined groups before closing a socket.
	/// When a socket is closed all multicast joins are dropped for you -- this
	/// routine just illustrates how to drop a group if you need to in the middle
	/// of the lifetime of a socket.
	/// </summary>
	public:
	void LeaveGroups()
	{
		try 
		{
			Console::WriteLine("Dropping membership to any joined groups...");
			for(int i = 0; i < multicastJoinList->Count ;i++)
			{
				for(int j = 0; j < localInterfaceList->Count ;j++)
				{
					// Create the MulticastOption structure which is required to drop the 
					//    multicast group (the same structure used to join the group is
					//    required to drop it).
					if ( mcastSocket->AddressFamily == AddressFamily::InterNetwork )
					{
						MulticastOption^     mcastOption = gcnew MulticastOption( 
							(IPAddress^) multicastJoinList[i], (IPAddress^) localInterfaceList[j]);
	            
						mcastSocket->SetSocketOption(
							SocketOptionLevel::IP,                    
							SocketOptionName::DropMembership,
							mcastOption
							);
					}
					else if ( mcastSocket->AddressFamily == AddressFamily::InterNetworkV6 )
					{
						IPv6MulticastOption^ ipv6McastOption = gcnew IPv6MulticastOption(
							(IPAddress^) multicastJoinList[i],
							( (IPAddress^) localInterfaceList[j] )->ScopeId
							);

						mcastSocket->SetSocketOption(
							SocketOptionLevel::IPv6,
							SocketOptionName::DropMembership,
							ipv6McastOption
							);
					}

					Console::WriteLine("Dropping multicast group {0} on interface {1}", 
						multicastJoinList[i]->ToString(),
						localInterfaceList[j]->ToString()
						);
				}
			}
		}
		catch(Exception^ err)
		{
			Console::WriteLine("LeaveGroups: No multicast groups joined: " + err->Message);
		}
	}

	/// <summary>
	/// This method sets the outgoing interface when a socket sends data to a multicast
	/// group. Because multicast addresses are not routable, the network stack simply
	/// picks the first interface in the routing table with a multicast route. In order
	/// to change this behavior, the MulticastInterface option can be used to set the
	/// local interface on which all outgoing multicast traffic is to be sent (for this
	/// socket only). This is done by converting the 4 byte IPv4 address (or 16 byte
	/// IPv6 address) into a byte array.
	/// </summary>
	/// <param name="sendInterface"></param>
	public:
	void SetSendInterface( IPAddress^ sendInterface )
	{
		// Set the outgoing multicast interface
		try
		{
			Console::WriteLine("Setting the outgoing multicast interface...");
			if ( mcastSocket->AddressFamily == AddressFamily::InterNetwork )
			{
				mcastSocket->SetSocketOption(
					SocketOptionLevel::IP,
					SocketOptionName::MulticastInterface,
					sendInterface->GetAddressBytes()
					);
	            
			}
			else
			{
				array<Byte>^ interfaceArray = BitConverter::GetBytes( (int)sendInterface->ScopeId );

				mcastSocket->SetSocketOption(
					SocketOptionLevel::IPv6,
					SocketOptionName::MulticastInterface,
					interfaceArray
					);
			}
			Console::WriteLine("Setting multicast send interface to: " + sendInterface->ToString());
		}
		catch ( SocketException^ err )
		{
			Console::WriteLine("SetSendInterface: Unable to set the multicast interface: {0}", err->Message);
			throw;
		}
	}     

	/// <summary>
	/// This method takes a string and repeatedly copies it into the send buffer
	/// to the length of the send buffer.
	/// </summary>
	/// <param name="message">String to copy into send buffer</param>
	public:
	void FormatBuffer( String^ message )
	{
		array<Byte>^  byteMessage = Text::Encoding::ASCII->GetBytes( message );
		int         index = 0;

		// First convert the string to bytes and then copy into send buffer
		Console::WriteLine("Formatting the string & copy to send buffer...");
		while ( index < dataBuffer->Length )
		{
			for(int j=0; j < byteMessage->Length ;j++)
			{

				dataBuffer[ index ] = byteMessage[ j ];
				index++;

				// Make sure we don't go past the send buffer length
				if ( index >= dataBuffer->Length )
				{
					break;
				}
			}
		}
	}
};


/// <summary>
/// This method prints usage information for multicast sample
/// </summary>
static void usage()
{
	Console::WriteLine("Executable_file_name [/i ip] [/b ip] [/m ip] [/p port] [/x num] [/t text]");
	Console::WriteLine("              [/n count] [/s | /r]");
    Console::WriteLine("Available options:");
    Console::WriteLine("   /i ip		Local interface to join multicast group(s) on");
    Console::WriteLine("   /b ip		Local IP to bind the socket to");
    Console::WriteLine("   /m ip		Multicast group to join");
    Console::WriteLine("   /p port		Port to listen on or send to");
    Console::WriteLine("   /x size		Length of send or receive buffer");
    Console::WriteLine("   /t text		Text message to put in send buffer");
    Console::WriteLine("   /n count		Number of times to send or receive");
    Console::WriteLine("   /s			Send data to a multicast group");
    Console::WriteLine("   /r			Receive data send to the group");
    Console::WriteLine();
}

/// <summary>
/// This is the main routine which parses the command line and creates the
/// multicast socket as given by the command line parameters. The sample
/// can either be invoked as a multicast sender or receiver.
/// 
/// To invoke as a multicast receiver:
///     multicast.exe /r /i 10.10.10.1 /m 234.6.6.6
/// To invoke as a multicast sender:
///     multicast.exe /s /i 10.10.10.99 /m 234.6.6.6
/// </summary>
/// <param name="appArguments">Command line arguments</param>
int main(array<System::String ^> ^args)
{
    MulticastEndpoint^ mcastEndpoint = gcnew MulticastEndpoint();
    String^ textMessage = "Hello Bastard World out there!";
    int port = 0, bufferLength = 1024, opCount = 10;
    bool isSender = true;

    // Parse the command line
    for (int i = 0; i < args->Length; i++)
    {
        try
        {
            if ((args[i][0] == '-') || (args[i][0] == '/'))
            {
				switch (Char::ToLower(args[i][1]))
                {

                    case 'b':
                        // Address to bind multicast socket to
						mcastEndpoint->bindAddress = IPAddress::Parse(args[++i]);
                        break;

                    case 'c':
                        // How many times to send or receive
						opCount = Convert::ToInt32(args[++i]->ToString());
                        break;

                    case 'i':
                        // Local interface to join group(s) on                                
						mcastEndpoint->localInterfaceList->Add(IPAddress::Parse(args[++i]));
                        break;

                    case 'm':
                        // Multicast group to join
						mcastEndpoint->multicastJoinList->Add(IPAddress::Parse(args[++i]));
                        break;

                    case 'p':
                        // Port number to bind to or send to
						port = Convert::ToInt32(args[++i]->ToString());
                        break;

                    case 'r':
                        // Application invoked as receiver
                        isSender = false;
                        break;

                    case 's':
                        // Application invoked as sender                    
                        isSender = true;
                        break;

                    case 't':
                        // Text message to send
                        textMessage = args[++i];
                        break;

                    case 'x':
                        // Length of send/receive buffer
						bufferLength = Convert::ToInt32(args[++i]->ToString());
                        break;

                    default:
                        usage();
                        return 0;
                }
            }
        }
        catch(Exception^ err)
        {
			Console::WriteLine("Error: " + err->Message);
            usage();
            return 0;
        }
    }

    // Make sure user specified at least on multicast group and interface to join
    if ((mcastEndpoint->multicastJoinList->Count == 0) || (mcastEndpoint->localInterfaceList->Count == 0))
    {
		Console::WriteLine("Please specify a multicast group and interface to join on!");
        usage();
        return 0;
    }

    // Now, create the multicast socket
	Console::WriteLine("Creating the multicast socket...");
    try
    {
        if (isSender == true)
        {
            // For the sender, we don't care what the local port the socket is bound to
            mcastEndpoint->Create(
                0,               // If the sender we don't care what local port we bind to 
                bufferLength
                );
        }
        else
        {
            mcastEndpoint->Create(port, bufferLength);
        }

        if (isSender == true)
        {
            IPEndPoint^ destinationGroup = gcnew IPEndPoint((IPAddress^)mcastEndpoint->multicastJoinList[0], port);
            int rc;

            // Set the send interface for all outgoing multicast packets
			Console::WriteLine("Setting the send interface for all outgoing multicast packets...");
            mcastEndpoint->SetSendInterface((IPAddress^)mcastEndpoint->localInterfaceList[0]);
            mcastEndpoint->FormatBuffer(textMessage);

            // Send the message the requested number of times
			Console::WriteLine("Sending the message the requested number of times...");
            for (int i = 0; i < opCount; i++)
            {
                try
                {
                    rc = mcastEndpoint->mcastSocket->SendTo(mcastEndpoint->dataBuffer, destinationGroup);
					Console::Write("Multicast SendTo() is OK. ");
					Console::WriteLine("Sent {0} bytes to {1}", rc, destinationGroup->ToString());
                }
                catch (SocketException^ err)
                {
					Console::WriteLine("Multiast SendTo() failed: " + err->Message);
                }
            }
        }
        else
        {
            IPEndPoint^ senderEndPoint = gcnew IPEndPoint((IPAddress^)mcastEndpoint->localInterfaceList[0], 0);
            EndPoint^ castSenderEndPoint = (EndPoint^)senderEndPoint;
            int rc;

            for (int i = 0; i < opCount; i++)
            {
                try
                {
                    rc = mcastEndpoint->mcastSocket->ReceiveFrom(mcastEndpoint->dataBuffer, castSenderEndPoint);
					Console::Write("Multicast ReceiveFrom() is OK...");
					// Setting to an instance of an object before using it
                    senderEndPoint = (IPEndPoint^)castSenderEndPoint;

					Console::WriteLine("Received {0} bytes from {1}: '{2}'\n",
                        rc,
                        senderEndPoint->ToString(),
						Text::Encoding::ASCII->GetString(mcastEndpoint->dataBuffer, 0, rc)
                        );
                }
                catch (SocketException^ err)
                {
					Console::WriteLine("Multicast ReceiveFrom() failed: " + err->Message);
                }
            }
        }

        // Drop membership to groups. This isn't required in this sample as closing the socket
        //    will implicitly leave all groups joined, but it's a good practice.
        mcastEndpoint->LeaveGroups();

    }
    catch(Exception^ err)
    {
		Console::WriteLine("An error occured creating the multicast socket: " + err->Message);

    }
    finally
    {
        if (mcastEndpoint->mcastSocket != nullptr)
        {
			Console::WriteLine("Closing the multicast socket...");
            mcastEndpoint->mcastSocket->Close();
        }
        mcastEndpoint = nullptr;
    }

    return 0;
}
